(* ==================================================================
The use of this Package is described in "Economic and Financial
Modeling with Mathematica", Hal Varian, editor, TELOS/Springer-Verlag,
1993, ISBN 0-387-97882-8.  $44.95 (Includes 1.44 MS-DOS disk.)  To
place order call 1-800-777-4643 or fax 1-201-348-4505.
=====================================================================*)

(* :Title: InfluenceDiagram 
*)
(* :Author:
	Robert J. Korsan 
*)
(* :Summary:

	Influence Diagrams in Mathematica

This package contains all the programs from the article, 'Decision 
Analytica' in the book "Econometrics and Finance in Mathematica" edited
by Dr. Hal Varian, Springer Verlag, The Electronic Library Of Science,
Santa Clara, CA.  ISBN 0-387-97882-8.


Any comments, bug reports, or requests should be forwarded to:
	Robert J. Korsan
	Decisions, Decisions!
	2136 Lyon Avenue
	Belmont, CA 94002-1639
    bobk@decide.com
    1 (415) 595-1206
*)
(* :Context: EconometricsAndFinance`DecisionAnalytica`InfluenceDiagram` 
*)
(* :Package Version: .1	(7/15/92 Alpha Release)
*)
(* :Copyright: Copyright 1992 by Robert J. Korsan

This package may be copied in its entirety for nonprofit purposes only.
Sale, other than for the direct cost of the media, is prohibited.  This
copyright notice must accompany all copies.

The author and Springer Verlag, make no representations, express or
implied, with respond to this documentation, of the software it
describes and contains, including without limitations, any implied
warranties of mechantability or fitness for a particular purpose, all
of which are expressly disclaimed.  The author and Springer Verlag, their
licensees, distributors and dealers shall in no event be liable for any
indirect, incidental, or consequential damages.
*)
(* :History:
	Version .1 by Robert J. Korsan, July 1992.
*)
(* :Keywords:
	influence diagram, relevance diagram, uncertainty, decision analysis
*)
(* :Source:
	Robert Korsan: 'Decision Analytica' in "Econometric and Finanace in
	        Mathematica", edited by Dr. Hal Varian,
			Springer Verlag
*)
(* :Mathematica Version: 2.0
*)

BeginPackage["DecisionAnalytica`InfluenceDiagram`",
             "DiscreteMath`Combinatorica`"]

AddConditioningTo::usage = "AddConditioningTo[id, e] is a function to
add conditioning information to the influence diagram object id. e is
the conditioning information. NB: A ValueNode object may be conditioned
by any other node object, but a ValueNode object may not condition any
of node object. Also, the node being conditioned must already be present
in the influence diagram object id.

The conditioning information, e, is given in one of the following forms:

(1) ValueComputedFrom[v, n]. v is a ValueNode object and n is a node or list
of nodes. (2) ChanceConditionedBy[c, n]. c is a ChanceNode object and n is a
node or list of nodes. (3) DecisionInformedBy[d, n]. d is a DecisionNode
object and n is a node or list of nodes. (4) ConsistencyRequiredBy[c, n]. 
c is a ConsistencyNode object and n is a node or list of nodes of any other
type."

Chance::usage = "Chance[label] is a function which creates a ChanceNode
object where label is a string used when displaying an Influence Diagram
containing this node. (In the future a ChanceNode object will contain
other quatities such as the probabililty distribution describing the
decision maker's uncertainty.)"

ChanceConditionedBy::usage = "ChanceConditionedBy is a head used to indicate
the conditioning information for a ChanceNode object."

Consistency::usage = "Consitency[label] is a function which creates a
ConsistencyNode object where label is a string used when displaying an
Influence Diagram containing this node. (In the future a ConsistencyNode
object will contain other quatities such as a input-output constraint
for the quantities entering and leaving the node.)"

ConsistencyRequiredBy::usage = "ConsistencyRequiredBy is a head used to
indicate the conditioning information for a ConsitencyNode object."

Decision::usage = "Decision[label] is a function which creates a DecisionNode
object where label is a string used when displaying an Influence Diagram
containing this node. (In the future a DecisionNode object will contain
other quantities such as the range of alternatives, whether optimization
means to minimize or maximize, etc.)"

DecisionInformedBy::usage = "DecisionInformedBy is a head used to indicate the
conditioning information for a DecisionNode object."

KnowledgeMap::usage = "KnowledgeMap is a header for an Influence Diagram
object. An Influence Diagram object is of the form KnowledgeMap[g,n]
where g is a Graph object corresponding to the Influence Diagram and n
is a list of the nodes in the Influence Diagram."

ShakeID::usage = "ShakeID[id] models the graph of id as a system of springs.
A minimum energy configuration is found and a KnowledgeMap object with
the transformed graph is returned. This function is useful for \"cleaning
up\" an influence diagram which has many arrow crossings."

ShowID::usage = "ShowID[id] is a function which displays a graphical
representation of id. id must be a KnowledgeMap object."

TrivialInfluenceDiagram::usage = "TrivialInfluenceDiagram[n] creates an
KnowledgeMap object containing just one node n where n is a ValueNode
object."

TrivialRelevanceDiagram::usage = "TrivialRelevanceDiagram[n] creates an
KnowledgeMap object containing just one node n where n is a ChanceNode
object."

TrivialKnowledgeMap::usage = "TrivialKnowledgeMap[n] creates an
KnowledgeMap object containing just one node n where n is a ChanceNode
object, a DecisionNode object, a ConsistencyNode object or a ValueNode
object."

Value::usage = "Value[label] is a function which creates a ValueNode 
object where label is a string used when displaying an Influence Diagram
containing this node. There may be only one ValueNode object in an
Influence Diagram and it may condition no other object. (In the future
a ValueNode object will contain other quatities such as a function to
compute the value of a state of nature given the quantities entering the
node.)"

ValueComputedFrom::usage = "ValueComputedFrom is a head used to indicate the
conditioning information for a ValueNode object."

InfluenceDiagram::notmember = "Warning: `1` is not a node in `2`."

InfluenceDiagram::notacyclic = "Adding an arrow from `1` to `2` causes
loops in `3`. Exiting."

PrependTo[$ContextPath,"DiscreteMath`Combinatorica`private`"]

Begin["`private`"]

Decision[l_String]:=DecisionNode[l]

Chance[l_String]:=ChanceNode[l]

Consistency[l_String]:=ConsistencyNode[l]

Value[l_String]:=ValueNode[l]

Decision[l_ColumnForm]:=DecisionNode[l]

Chance[l_ColumnForm]:=ChanceNode[l]

Consistency[l_ColumnForm]:=ConsistencyNode[l]

Value[l_ColumnForm]:=ValueNode[l]

TrivialInfluenceDiagram[v_ValueNode]:=
 Module[{IDgraph,IDnodes},
  IDgraph=EmptyGraph[1];
  IDnodes={v};
  KnowledgeMap[IDgraph,IDnodes]
 ]
 
TrivialRelevanceDiagram[v_ChanceNode]:=
 Module[{IDgraph,IDnodes},
  IDgraph=EmptyGraph[1];
  IDnodes={v};
  KnowledgeMap[IDgraph,IDnodes]
 ]

TrivialKnowledgeMap[v_]:=
 Module[{IDgraph,IDnodes},
  IDgraph=EmptyGraph[1];
  IDnodes={v};
  KnowledgeMap[IDgraph,IDnodes]
 ]/; MemberQ[{DecisionNode,
              ChanceNode,
              ConsistencyNode,
              ValueNode},Head[v]]

AddNewNodes[Old_List,{}]:=Old

AddNewNodes[Old_List,{a_,b___}]:=
  AddNewNodes[AddNewNodes[Old,a],{b}]

AddNewNodes[Old_List,x_]:=
 Module[{old=Old},
  If[MemberQ[old,x],old,Append[old,x]]] /; Not[MatchQ[Head[x],List]]

SetAttributes[AddConditioningTo,HoldFirst]

AddConditioningTo[id_Symbol,
            ChanceConditionedBy[d_ChanceNode,n_]]:=
  AddConditioningTo[Unevaluated[id],DirectedEdge[d,n]] /; NodeOrListOfNodesNotValueQ[n]

AddConditioningTo[id_Symbol,
            ConsistencyRequiredBy[d_ConsistencyNode,n_]]:=
  AddConditioningTo[Unevaluated[id],DirectedEdge[d,n]] /; NodeOrListOfNodesNotValueQ[n]

AddConditioningTo[id_Symbol,
            DecisionInformedBy[d_DecisionNode,n_]]:=
  AddConditioningTo[Unevaluated[id],DirectedEdge[d,n]] /; NodeOrListOfNodesNotValueQ[n]

AddConditioningTo[id_Symbol,
            ValueComputedFrom[d_ValueNode, n_]]:=
  AddConditioningTo[Unevaluated[id],DirectedEdge[d, n]] /; NodeOrListOfNodesNotValueQ[n]

AddConditioningTo[id_Symbol,DirectedEdges_List]:=
Module[{IDgraph=First[id], IDnodes=Last[id], DElist=DirectedEdges,
        idlabel=ToString[Unevaluated[id]],
        EdgesTransposed=Transpose[DirectedEdges],
        target, FromNodes, UpdatedNodes, NewNodes,
        ArrowFrom,x,y},
  (*
  *  If the node which is supposed to be conditioned is
  *  not a member of the influence diagram, print a warning
  *  message and add it to the UpdatedNodes list.
  *)
  
  target=First[Last[EdgesTransposed]];
  If[!MemberQ[IDnodes,target],
    Message[InfluenceDiagram::notmember, target, idlabel];
    y=1+V[IDgraph],
    y=First[Flatten[Position[IDnodes,target]]]
  ];
  
  (*
  *  Now lets get the list of nodes with (possibly new ones to be added)
  *  and add vertices to the graph until we have enough.
  *)
  FromNodes=First[EdgesTransposed]; PrependTo[FromNodes,target];
  
  UpdatedNodes=AddNewNodes[IDnodes,FromNodes];
  (* NB: Order is important in this list. *)
  While[V[IDgraph]!=Length[UpdatedNodes], IDgraph=AddVertex[IDgraph]];
  
  (*
  *  Now that we have enough vertices, lets add in the edges from
  *  the list given. Make sure that the graph remains acyclic. Print
  *  an error message and return without changing anything if this is
  *  not the case.
  *)
  cnt=0;
  While[0!=Length[DElist], 
    ArrowFrom=First[First[DElist]]; 
    x=Position[UpdatedNodes,ArrowFrom]; 
    x=First[First[x]]; 
    IDgraph=AddEdge[IDgraph,{x,y},Directed];
    If[!AcyclicQ[IDgraph],
      Message[InfluenceDiagram::notacyclic, ArrowFrom, target, idlabel];
      Return[]];
    
   DElist=Drop[DElist,1]];
   
 (*
 *  Now that the edges and Vertices are in, lets shake the spring
 *  embedding of the graph and replace the vertices. Then we are
 *  done.
 *)
 
 target=Map[Head,UpdatedNodes];
 If[V[IDgraph]>3,
    If[MemberQ[target,ValueNode],
      IDgraph=ChangeVertices[IDgraph,
            Insert[CircularVertices[V[IDgraph]-1],{0,0},
              First[Flatten[Position[target,ValueNode]]]]],
      IDgraph=ChangeVertices[IDgraph,
            Insert[CircularVertices[V[IDgraph]-1],{0,0},1]]
    ],
    IDgraph=ChangeVertices[IDgraph,
            CircularVertices[V[IDgraph]]]
];
 
 id=KnowledgeMap[IDgraph,UpdatedNodes] 
  
]

DirectedEdge[ArrowTo_,ArrowFrom_]:=
  Flatten[
   Outer[List, Flatten[{ArrowFrom}], {ArrowTo}], 1]

NodeOrListOfNodesNotValueQ[n_]:=
 Module[{n1=Map[Head,Flatten[{n}]]},
  3 == Length[ Union[{ChanceNode,ConsistencyNode,DecisionNode}, n1] ]
 ]

IDFindPlotRange[v_List] :=
	Module[{xmin=Min[Map[First,v]], xmax=Max[Map[First,v]],
			ymin=Min[Map[Last,v]], ymax=Max[Map[Last,v]]},
		{ {xmin - 0.20 Max[1,xmax-xmin], xmax + 0.10 Max[1,xmax-xmin]},
		  {ymin - 0.20 Max[1,ymax-ymin], ymax + 0.05 Max[1,ymax-ymin]} }
	]

IDPointsAndLines[Graph[e_List,v_List],NodeType_List] :=
	Module[{i,pairs=ToOrderedPairs[Graph[e,v]],
	        operators={
			  {GrayLevel[0.7], Disk[#,0.03]}&,
			  {GrayLevel[0.7], Rectangle[#-{0.03,0.03},#+{0.03,0.03}]}&,
			  {Circle[#,.04],Circle[#,.03]}&,
			  {GrayLevel[0.3], Polygon[Table[{Cos[(2i-3)Pi/8]/10, Sin[(2i-3)Pi/8]/10}0.5+#, {i,1,8}]]}&
			  }[[ NodeType ]]
			 },
		Join[
			Table[(operators[[i]])[ v[[i]] ], {i,Length[NodeType]}],
			Map[(Line[Chop[ v[[#]] ]])&,pairs]
		]
	]

IDArrows[Graph[e_,v_]] :=
	Module[{pairs=ToOrderedPairs[Graph[e,v]],
	        size, triangle},
		size = Min[0.04, MinimumEdgeLength[v,pairs]/3];
		triangle={ {0,0}, {-size,size/2}, {-size,-size/2} };
		
		Map[
			(Polygon[
				TranslateVertices[
					RotateVertices[triangle,
						Arctan[Apply[Subtract,v[[#]]]]+Pi],
					(0.7v[[ #[[2]] ]]+0.3v[[ #[[1]] ]])
				]
			])&,
			pairs
		]
	]
	
IDGraphLabels[v_List,l_List] :=
	Module[{i},
		Table[ Text[ l[[i]],v[[i]]-{0.,0.1},{0,1} ],{i,Length[v]}]
	]

ShowID[id_KnowledgeMap, opts___] :=
	Module[{arrows, g, pairs, IDgraph=First[id],
	        IDnodes=Last[id], labels,
	        NodeTypes, ptslines, range, standard, v},
	    pairs = ToOrderedPairs[IDgraph];
	    g = NormalizeVertices[IDgraph];
		v = Vertices[g];
		labels = Flatten[Map[List@@#&,IDnodes]];
		labels = IDGraphLabels[v,labels];
		standard={ChanceNode, DecisionNode,ConsistencyNode,ValueNode};
		
		NodeTypes=Flatten[Map[Position[standard,#]&,Map[Head,IDnodes]]];
		ptslines=IDPointsAndLines[g,NodeTypes];
        arrows=IDArrows[g];
        range=IDFindPlotRange[v];

		Show[
			Graphics[
				Join[
					ptslines,
					arrows,
					labels
				]
			],AspectRatio->Automatic, PlotRange->range, opts 
		]
	]
	
IDCalculateForce[u_Integer,g_Graph,em_List] :=
	Module[{n=V[g],stc=0.25,gr=10.0,e=Edges[g],f={0.0,0.0},spl=10.0,v,dsquared},
		Do [
			dsquared = Max[0.001, Apply[Plus,(em[[u]]-em[[v]])^2] ];
			f += (1-e[[u,v]]) (gr/dsquared) (em[[u]]-em[[v]])
				- e[[u,v]] stc Log[dsquared/spl] (em[[u]]-em[[v]]),
			{v,n}
		];
		f
	]

IDSpringEmbedding[g_Graph,step_:10,inc_:0.25] :=
	Module[{new=old=Vertices[g],n=V[g],i,u,g1=MakeUndirected[g]},
		Do [
			Do [
				new[[u]] = old[[u]]+inc*IDCalculateForce[u,g1,old],
				{u,n}
			];
			old = new,
			{i,step}
		];
		Graph[Edges[g],new]
	]

ShakeID[id_KnowledgeMap,step_:10,inc_:0.25]:=
 Module[{g=First[id]},
   
   g=IDSpringEmbedding[ShakeGraph[g,0.5], step, inc];
   g=ChangeVertices[g, -1+2*NormalizeVertices[Vertices[g]]];
   KnowledgeMap[ g  ,Last[id] ]
 ]
   

End[]
EndPackage[ ]


