(* ==================================================================
The use of this Package is described in "Economic and Financial
Modeling with Mathematica", Hal Varian, editor, TELOS/Springer-Verlag,
1993, ISBN 0-387-97882-8.  $44.95 (Includes 1.44 MS-DOS disk.)  To
place order call 1-800-777-4643 or fax 1-201-348-4505.
=====================================================================*)

BeginPackage["Aim`"]
(*
Mathematica code for implementing
Anderson Moore technique for solving linear discrete time models.

Gary Anderson
July, 1989

See
A Linear Algebraic Approach for Solving Linear Perfect Foresight Models
by Anderson and Moore

*)

SymbolicAim::usage=
"   SymbolicAim[SymbolicHMatrix]
      Returns {AuxiliaryInitialConditions,Eigenvalues,{},
	stateSpaceTransitionMatrix,TransformedHMatrix}"
SymbolicAimWithEigenvectors::usage=
"   SymbolicAimWithEigenvectors[SymbolicHMatrix]
      Returns {AuxiliaryInitialConditions,Eigenvalues,Eigenvectors,
	stateSpaceTransitionMatrix,TransformedHMatrix}"
NumericalAim::usage=
"   NumericalAim[NumericalHMatrix]
      Returns {AuxiliaryInitialConditions,Eigenvalues,Eigenvectors,
	stateSpaceTransitionMatrix,TransformedHMatrix}"
NumericalComputeQMatrix::usage=
"   NumericalComputeQ[NumericalHMatrix]
	Returns QMatrix"

AugmentQMatrix::usage=
"   AugmentQMatrix[Eigenvalues,Eigenvectors,AuxiliaryInitialConditions]
	Returns QMatrix"
ComputeBMatrix::usage=
"   ComputeBMatrix[QMatrix] Returns BMatrix"

ReverseSymbolicAim::usage=
"   ReverseSymbolicAim[SymbolicHMatrix]
      Returns {AuxiliaryInitialConditions,Eigenvalues,{},
	stateSpaceTransitionMatrix,TransformedHMatrix}"
ReverseNumericalAim::usage=
"   ReverseNumericalAim[NumericalHMatrix]
      Returns {AuxiliaryInitialConditions,Eigenvalues,Eigenvectors,
	stateSpaceTransitionMatrix,TransformedHMatrix}"

derivativesOfVectorFunction::usage=
"derivativesOfVectorFunction[functions_,variables_]
   Returns MatrixOfDerivatives"

makeSymbolicAsymptoticHMatrix::usage=
"makeSymbolicAsymptoticHMatrix::usage[model_,periodicity_]
  Returns asymptotic H Matrix "
stateSpaceTransitionMatrix::usage=
"stateSpaceTransitionMatrix[H_]
Returns {TransitionMatrix,AuxiliaryInitialConditions,TransformedHMatrix}"

eliminateInessentialLags::usage=
"eliminateInessentialLags[{stateSpaceTransitionMatrix_,VariablesList_}]
Returns[{MinimalDimensionStateSpaceTransitionMatrix,
			MinimalStateSpaceVariablesList}]"
CombineGandH::usage:=
"CombineGandH[gmat_,hmat_] Returns sum of two matrix, pads g to
conform with h"

GetModelFunctions::usage="Get the functions for the model
\n returns a pure function which represents the implicit function
\n********************************
\n amodel is of form {function,variableList,parameterList,{neq,nlag,nlead}}

GetModelFunctions[amodel]"
GetModelVariableList::usage="GetModelVariableList[amodel]"
GetModelParameterList::usage="GetModelParameterList[amodel]"
GetModelMaximumLag::usage="GetModelMaximumLag[amodel]"
GetModelMaximumLead::usage="GetModelMaximumLead[amodel]"
GetModelNumberOfEquations::usage="GetModelNumberOfEquations[amodel]"

Begin["private`"]
GetModelFunctions[amodel_]:=First[amodel];
GetModelVariableList[amodel_]:=First[Rest[amodel]];
GetModelParameterList[amodel_]:=First[Rest[Rest[amodel]]];
GetModelMaximumLag[amodel_]:=First[Last[amodel]];
GetModelMaximumLead[amodel_]:=Last[Last[amodel]];
GetModelNumberOfEquations[amodel_]:=Length[First[Rest[amodel]]];



(*Compute the state vector at time t*)
stateVector[amodel_,t_]:=Block[{},
	Map[Function[x,Apply[x,{t}]],GetModelVariableList[amodel]]];


(*
In order to compute solutions for equation 1 we will need to construct 
systems of equations using the model system of equations evaluated at
different points in time.
*)

TimeTArgument[t_,amodel_]:=Block[{lag,maxlag,maxlead},
	maxlag=GetModelMaximumLag[amodel];maxlead=GetModelMaximumLead[amodel];
	Return[Flatten[Table[
			Map[
			Function[x,Apply[x,{t+lag}]],
				GetModelVariableList[amodel]],
				{lag,-maxlag,maxlead,1}]]]];
(* T and offset must be actual numbers not symbols *)




(* T and offset must be actual numbers not symbols *)
Tfunctions[T_,amodel_,offset_]:=
	Block[{fns,t,fn},
		If[T==0,Return[{}],
			fn=GetModelFunctions[amodel];
			fns=Table[Apply[fn,TimeTArgument[t+offset,amodel]],
						{t,0,T-1}]];
			Return[Flatten[fns]]];


derivativesOfVectorFunction[functions_,variables_]:=


	Array[Function[{i,j},D[functions[[i]],variables[[j]]]],
					{Length[functions],Length[variables]}]

makeSymbolicPeriodicPoints[amodel_,period_]:=
	Block[{pathvector,limits,maxlag,maxlead,numberofequations},
	maxlag=GetModelMaximumLag[amodel];
	maxlead=GetModelMaximumLead[amodel];
	numberofequations=GetModelNumberOfEquations[amodel];
	pathvector=Flatten[Table[stateVector[amodel,n],
				{n,-maxlag,maxlead+period-1}]];
	Return[Take[Map[ ToExpression[StringJoin["asymptotic",ToString[#]]] &, 
			pathvector],numberofequations period]]];

cycleMap[t_,period_]:=Block[{remainder},
	remainder=Mod[t,period];
	If[remainder==0,Return[period],Return[remainder]]];

makeSymbolicAsymptoticRestrictions[amodel_,period_]:=
	Block[{maxlag,maxlead,numberofequations,pathvector,i,cycle},
	maxlag=GetModelMaximumLag[amodel];
	maxlead=GetModelMaximumLead[amodel];
	numberofequations=GetModelNumberOfEquations[amodel];
	pathvector=Flatten[Table[stateVector[amodel,n],
				{n,-maxlag,maxlead+period-1}]];
	cycle=makeSymbolicPeriodicPoints[amodel,period];
		Return[Map[#[[1]]->#[[2]] &,
				Table[
				{pathvector[[i]],
				cycle[[cycleMap[i,numberofequations period]]]},
				{i,Length[pathvector]}]]]];


makeVectorFunctionDerivatives[functions_,variables_]:=
	Array[Function[{i,j},D[functions[[i]],variables[[j]]]],
					{Length[functions],Length[variables]}]

matrixofDerivatives[offset_,t_,amodel_]:=
	Block[{i,fns,vbls,maxlead,maxlag},
                maxlag=GetModelMaximumLag[amodel];maxlead=GetModelMaximumLead[amodel];
                numbeq=GetModelNumberOfEquations[amodel];
		If[t==0,Return[{}],
        		variables=Flatten[Table[
                                stateVector[amodel,i-maxlag-1+offset],
        			{i,maxlag+maxlead+t+1}]];
			fns=Tfunctions[t,amodel,offset];
			Return[Array[
				Function[{i,j},
				D[fns[[i]],variables[[j]]]],
					{numbeq t,
                                                numbeq (t+maxlead+maxlag)}]]]];



makeSymbolicAsymptoticHMatrix[amodel_,period_]:=
        Block[{pathvector,maxlag,maxlead,restrictions},
	maxlag=GetModelMaximumLag[amodel];
	maxlead=GetModelMaximumLead[amodel];
	pathvector=Flatten[Table[stateVector[amodel,n],
				{n,-maxlag,maxlead+period-1}]];
        restrictions=makeSymbolicAsymptoticRestrictions[amodel,period];
	Return[makeVectorFunctionDerivatives[
			Tfunctions[period,amodel,0],pathvector]/.
                                restrictions]]


makeNonSingular[H_]:=
	Block[
		{i,zappingStuff,zappingValues,zappingMatrix,NewH,LclH,Qmat,
		zeroRows,gtheta,tau,theta,htheta,
                hthetacols,whoops,numberEquations},
		numberEquations=Dimensions[H][[1]];
		numberBlocks=(Dimensions[H][[2]])/numberEquations;
                hthetacols=Table[i,
                        {i,numberEquations (numberBlocks -1)+1,
                                numberEquations numberBlocks}];
		LclH=H;
(*Note colon assignment so evaluated new each time referenced*)
		htheta:=Transpose[Transpose[LclH][[hthetacols]]];
		Qmat={};
		While[Det[htheta]==0,
zappingMatrix=makeZappingMatrix[htheta];
		NewH=(zappingMatrix . LclH);
		zerorows=FindZeroRows[Transpose[
			Transpose[NewH][[hthetacols]]]];
	Qmat=Join[Qmat,NewH[[zerorows]]];
		LclH=shift[zerorows,NewH]
		];
		NewH=-Inverse[htheta] . LclH;
                Clear[i];
                If[Qmat !={},Qmat=Transpose[Transpose[Qmat][[Table[i,
                                        {i,1,numberEquations numberBlocks}]]]],];
		Return[{NewH,Qmat}]]

	isZero[x_]:=If[NumberQ[x],If[x==0,True,False],False];
	

makeZappingMatrix[amat_]:=Block[{nullspace,nullcol,nullrow,lastrows,goodcols},
	nullspace=RowReduce[NullSpace[Transpose[amat]]];
	{nullrow,nullcol}=Dimensions[nullspace];
	goodcols=Map[Min[Flatten[Position[#,1]]]&,nullspace];
	lastrows=IdentityMatrix[nullcol][[Complement[Range[1,Length[amat]],
			goodcols]]];
	Return[Join[nullspace,lastrows]]];

makeZappingMatrix2[amat_]:=Block[{nullspace,nullcol,nullrow,lastrows},
	nullspace=RowReduce[NullSpace[Transpose[amat]]];
	{nullrow,nullcol}=Dimensions[nullspace];
	lastrows=IdentityMatrix[nullcol][[FindZeroColumns[nullspace]]];
	Return[Join[nullspace,lastrows]]];


	shift[rows_,H_]:=
		Block[
			{tfrows,numberBlocks,NewH,AnH,numberEquations,i},
		numberEquations=Dimensions[H][[1]];
		numberBlocks=(Dimensions[H][[2]])/numberEquations;
	tfrows=Map[MemberQ[rows,#]&,Range[1, numberEquations]];
			NewH={};
			AnH={};
			For[j=1,j<=numberEquations,j++,
                                If[tfrows[[j]],
                                        AppendTo[NewH,
                                                RotateRight[H[[j]],
                                                        numberEquations]],
                                        AppendTo[NewH,H[[j]]]]];
			Return[NewH]
			]
			

stateSpaceTransitionMatrix[H_]:=
	Block[{i,Amat,dimension,numberBlocks,newh,imat,
			numberLeads,numberEquations},
		numberEquations=Dimensions[H][[1]];
		numberBlocks=(Dimensions[H][[2]])/numberEquations;
		dimension=numberEquations ( numberBlocks - 1 );
		imat=IdentityMatrix[dimension];
		{newh,qmat}=makeNonSingular[H];
		For[i=1,i<=dimension,i++,
		For[j=1,j<=dimension,j++,
			If[i>dimension-numberEquations,
				fillmat[i,j]=
				newh[[i+numberEquations
					-dimension]][[j]],
			If[j<=numberEquations,
				fillmat[i,j]=0,
				fillmat[i,j]=
				imat[[i]][[j-numberEquations]]]]]];
		Amat=Array[fillmat,
			{dimension,dimension}];
If[qmat != {},
		qmat=Transpose[Take[Transpose[qmat],Dimensions[Amat][[1]]]],];
		Return[{Amat,qmat,newh}]];


aim[H_]:=
	Block[{i,newh,qmat,theamat,evecs,evals},
		{theamat,qmat,newh}=stateSpaceTransitionMatrix[H];
		{evals,evecs}=Eigensystem[Transpose[theamat]];
		Return[{qmat,evals,evecs,theamat,newh}]]

SymbolicAim[H_]:=
	Block[{i,newh,qmat,theamat,evecs,evals},
		{theamat,qmat,newh}=stateSpaceTransitionMatrix[H];
		evals=Eigenvalues[Transpose[theamat]];
		Return[{qmat,evals,{},theamat,newh}]]

SymbolicAimWithEigenvectors[H_]:=
	Block[{i,newh,qmat,theamat,evecs,evals},
		{theamat,qmat,newh}=stateSpaceTransitionMatrix[H];
		{evals,evecs}=Eigensystem[Transpose[theamat]];
		Return[{qmat,evals,evecs,theamat,newh}]]

NumericalAim[H_]:=
	Block[{i,newh,qmat,theamat,evecs,evals},
		{theamat,qmat,newh}=stateSpaceTransitionMatrix[H];
		{evals,evecs}=Eigensystem[Transpose[N[theamat]]];
		Return[{qmat,evals,evecs,theamat,newh}]]



augmentQ[roots_,Q_,EVs_]:=
        If[roots=={},Q,
	If[Abs[First[roots]]<=1,
                augmentQ[Rest[roots],Q,Rest[EVs]],
                Append[augmentQ[Rest[roots],Q,Rest[EVs]],
                        First[EVs]]]];

getq[H_]:=Block[{evs,evals,q,a,newh},
                        {q,evals,evs,a,newh}=aim[H];
                        Return[augmentQ[evals,q,evs]]]

NumericalComputeQMatrix[H_]:=Block[{evs,evals,q,a,newh},
                        {q,evals,evs,a,newh}=NumericalAim[H];
                        Return[augmentQ[evals,q,evs]]]

ReverseGroup[length_,list_]:=Flatten[Reverse[Partition[list,length]]];

reverseaim[H_]:=
	Block[{i,newh,qmat,theamat,evecs,evals,numbeq},
		numbeq=Length[H];
		{theamat,qmat,newh}=stateSpaceTransitionMatrix[
				ReverseGroup[numbeq,#]&/@H];
		{evals,evecs}=Eigensystem[Transpose[theamat]];
		Return[{qmat,evals,evecs,theamat,newh}]]

ReverseSymbolicAim[H_]:=
	Block[{i,newh,qmat,theamat,evecs,evals,numbeq},
			numbeq=Length[H];
		{theamat,qmat,newh}=stateSpaceTransitionMatrix[
				ReverseGroup[numbeq,#]&/@H];
		evals=Eigenvalues[Transpose[theamat]];
		Return[{qmat,evals,{},theamat,newh}]]

ReverseNumericalAim[H_]:=
	Block[{i,newh,qmat,theamat,evecs,evals,numbeq},
		numbeq=Length[H];
		{theamat,qmat,newh}=stateSpaceTransitionMatrix[
				ReverseGroup[numbeq,#]&/@H];
		{evals,evecs}=Eigensystem[Transpose[N[theamat]]];
		Return[{qmat,evals,evecs,theamat,newh}]]

ComputeBMatrix[QMatrix_]:=Block[{QR,Qrows},
	Qrows=Length[QMatrix];Qcols=Length[QMatrix[[1]]];
	QR=Transpose[Take[Transpose[QMatrix],-Qrows]];
        Return[(-Inverse[QR] .
		QMatrix)[[Range[1,Qrows],Range[1,Qcols-Qrows]]]]]


kron[a_,b_]:=Block[{aidim,ajdim,bidim,bjdim,ipt,jpt},
{aidim,ajdim}=Dimensions[a];
{bidim,bjdim}=Dimensions[b];
Return[Table[a[[Quotient[ipt-1,bidim]+1]][[Quotient[jpt-1,bjdim]+1]]
			b[[Mod[ipt-1,bidim]+1]][[Mod[jpt-1,bjdim]+1]],
			{ipt,aidim bidim},{jpt,ajdim bjdim}]]];

eliminateInessentialLags[AMatrixVariableListPair_]:=
	Block[{firstzerocolumn,matrixsize},
	matrixsize=Length[AMatrixVariableListPair[[1]]];
	firstzerocolumn=FindFirstZeroColumn[
			AMatrixVariableListPair[[1]]];
	Return[
		If[IntegerQ[firstzerocolumn],
		eliminateInessentialLags[
			List[
			((AMatrixVariableListPair[[1]])[[
				Drop[Range[1,matrixsize],{firstzerocolumn}],
				Drop[Range[1,matrixsize],{firstzerocolumn}]]]),
			Drop[AMatrixVariableListPair[[2]],{firstzerocolumn}]]],
		AMatrixVariableListPair]]]
FindFirstZeroColumn[AMatrix_]:=Block[{columnsumofabs},
	columnsumofabs=Map[Apply[And,#]&,
		Map[Map[# == 0&,#]& , Transpose[AMatrix]]];
	Return[Apply[Min,Position[columnsumofabs,True]]]]

FindZeroColumns[AMatrix_]:=Block[{columnsumofabs},
	columnsumofabs=Map[Apply[And,#]&,
		Map[Map[# == 0&,#]& , Transpose[AMatrix]]];
	Return[Flatten[Position[columnsumofabs,True]]]]
FindZeroRows[AMatrix_]:=Block[{rowsumofabs},
	rowsumofabs=Map[Apply[And,#]&,
		Map[Map[# == 0&,#]& , AMatrix]];
	Return[Flatten[Position[rowsumofabs,True]]]]

stableStateSpaceTransitionMatrix[B_]:=Join[
	Take[IdentityMatrix[Length[B[[1]]]],Range[Length[B]+1,Length[B[[1]]]]],
	B];

NForecasts[B_,Path_,N_]:=Block[{tmp},
	If[N == 0 , Return[Path],
	Return[tmp=NForecasts[B,Path,N-1];
		Join[tmp,
			 B . Take[ tmp, 
				-Length[B[[1]]]]]]]];

CombineGandH[gmat_,hmat_]:=Block[
	{fmat,hcols,gcols},
	gcols=Length[gmat[[1]]];hcols=Length[hmat[[1]]];
	fmat= addZeroColsAtEnd[gmat,hcols-gcols] + hmat];
	
	
addZeroColsAtEnd[amat_,numberOfColsToAdd_]:=
	Map[Join[#,Table[0,{numberOfColsToAdd}]]&,amat];
addZeroColsAtBeginning[amat_,numberOfColsToAdd_]:=
	Map[Join[Table[0,{numberOfColsToAdd}],#]&,amat];
addZeroRowsAtEnd[amat_,numberOfRowsToAdd_]:=
	Join[amat,Table[Table[0,{Length[amat[[1]]]}],{numberOfRowsToAdd}]];
addZeroRowsAtBeginning[amat_,numberOfRowsToAdd_]:=
	Join[Table[Table[0,{Length[amat[[1]]]}],{numberOfRowsToAdd}],amat];
addIdentityAtRight[amat_]:=Block[{norows,imat,res},
	norows=Length[amat];imat=IdentityMatrix[norows];
	Return[Table[Join[amat[[i]],imat[[i]]],{i,norows}]]];
addIdentityAtBottom[amat_]:=Block[{nocols,imat,res},
	nocols=Length[amat[[1]]];imat=IdentityMatrix[nocols];
	Return[Join[amat,imat]]]


epsilonTMatrix[gmat_,hmat_,bmat_,timeLag_]:=Block[
	{tau,tauPlusTheta,forecastsneeded,forecastpart,numbVariables,datapart},
	numbVariables=Length[bmat];
	tau=Length[bmat[[1]]]/numbVariables;
	tauPlusTheta=(Length[hmat[[1]]]/numbVariables) -1;
	forecastsneeded=tauPlusTheta-tau  + timeLag;
	forecastpart=Which[timeLag == 1,
				addZeroColsAtEnd[
				hmat .
				NForecasts[bmat,
				IdentityMatrix[tau numbVariables],
				forecastsneeded], numbVariables],
			timeLag == 0,
				addZeroColsAtBeginning[
				hmat .
				addZeroRowsAtBeginning[
				NForecasts[bmat,
				IdentityMatrix[tau numbVariables],
				forecastsneeded], numbVariables], numbVariables],
			timeLag >1,
				addZeroColsAtEnd[
				addZeroColsAtBeginning[ hmat,
					numbVariables (timeLag-1)] .
				NForecasts[bmat,
				IdentityMatrix[tau numbVariables],
				forecastsneeded],timeLag
					numbVariables]];
		datapart=Which[timeLag<=1,
					gmat,
				timeLag>1,
					addZeroColsAtBeginning[gmat,
						(timeLag - 1)
						numbVariables]];
	Return[datapart+forecastpart]];

		

VarForm[gmat_,hmat_,bmat_,timeLag_]:=Block[
	{epsmat,jacobian,epsrows},
	epsmat=epsilonTMatrix[gmat,hmat,bmat,timeLag];
	epsrows=Length[epsmat];
	jacobian=Transpose[Take[Transpose[epsmat],-epsrows]];
	Return[{jacobian, -Inverse[jacobian] . epsmat}]]
End[]
EndPackage[]
